;;;-
;;; Copyright (c) 2004 AIAN
;;; All rights reserved.
;;;
;;; Redistribution and use in source and binary forms, with or without
;;; modification, are permitted provided that the following conditions
;;; are met:
;;;
;;; 1. Redistributions of source code must retain the above copyright
;;;    notice, this list of conditions and the following disclaimer. 
;;; 2. Redistributions in binary form must reproduce the above copyright 
;;;    notice, this list of conditions and the following disclaimer in the
;;;    documentation and/or other materials provided with the distribution.
;;;
;;; THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND ANY
;;; EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
;;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
;;; PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR
;;; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
;;; EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
;;; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
;;; PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
;;; OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;;;
;;; $Id:$

;;; ============================================================================
;;; BASIL - game ROM reader writer
;;; ============================================================================
;;; Reader instruction
;;; 
;;; 'u'                      .... reset
;;; 'a <adr_high> <adr_low>' .... address set
;;; 'r'                      .... read data from cart
;;; 'p'                      .... read ROM with incremental addressing
;;; 'b'                      .... read ROM bulk
;;; 'w <data>'               .... write data to cart
;;; ============================================================================

.include "8515def.inc"

.equ    VERSION         = $02   ; version
                                ; (upper 4bit is major, lower is minor)
;.equ    BR              = 12    ; set 38400bps @8MHz XTAL
.equ    BR              = 3     ; set 115200bps @7.3728MHz XTAL
;;; PORTD
.equ    RX              = 0     ; - no use -
.equ    TX              = 1     ; - no use -
.equ    CTS             = 2     ; RS232C CTS (INPUT)
.equ    RTS             = 3     ; RS232C RTS (OUTPUT)
.equ    WR              = 4     ; cart ~WR
.equ    RD              = 5     ; cart ~RD
.equ    CS              = 6     ; cart ~CS
.equ    RES             = 7     ; cart ~RES
;;; SEND ENABLED FLAG
.equ    CTS_SEND_OK     = $0f   ; CTS enabled
.equ    BASIL_SEND_OK   = $f0   ; `send_complete routine was calld' flag
.equ    SEND_ENABLED    = $ff   ; send enabled
.equ    SEND_DISABLED   = $00   ; send disabled

;;; Registors
.def    temp            = r16
.def    status          = r17   ; SREG preservation
.def    f_data_exist    = r18   ; unprocessed data exit flag
.def    f_send_enabled  = r19   ; send enabled flag
.def    rx_data         = r20   ; UART Rx data
.def    tx_data         = r21   ; UART Tx data
.def    b_counter       = r22   ; bulk read counter (used in 'b' command)
.def    t_counter       = r23   ; timer counter
.def    address_low     = r24   ; address high
.def    address_high    = r25   ; address low

.cseg
.org 0
        rjmp    reset                   ; reset
        reti                            ; external interrupt 0
        reti                            ; external interrupt 1
        reti                            ; timer/counter1 capture event
        reti                            ; timer/counter1 compare match A
        reti                            ; timer/counter1 compare match B
        reti                            ; timer/counter1 overflow
        reti                            ; timer/counter0 overflow
        reti                            ; serial tranfer complete
        rjmp    receive_complete        ; UART RX complete
        rjmp    send_enable             ; UART data register empty
        reti                            ; UART TX complete
        reti                            ; analog comparator

;;;======================================================================
;;; Reset
;;;======================================================================        
reset:
        cli                     ; disable global interruption
        ;; Stack init
        ldi     temp, high(RAMEND)
        out     SPH, temp
        ldi     temp, low(RAMEND)
        out     SPL, temp
        ;; PORT init
        ldi     temp, $ff       ; PORTA -- all output
        out     DDRA, temp
        ldi     temp, $00       ; PORTB -- all input (8bit data bus)
        out     DDRB, temp
        ldi     temp, $ff       ; PORTB pull up
        out     PORTB, temp
        ldi     temp, $ff       ; PORTC -- all output
        out     DDRC, temp
        ldi     temp, $fa       ; PORTD -- all output but PD0(RxD), PD2(CTS)
        out     DDRD, temp
        ldi     temp, $ff
        out     PORTB, temp
        ;; UART init
        ldi     temp, BR        ; set baud rate
        out     UBRR, temp
        ldi     temp, 0b11111000; enable RxD, TxD and their interruption
        out     UCR, temp
        sei                     ; enable global interruption
        ;; cart init
        cbi     PORTD, RES
        rcall   wait
        sbi     PORTD, RES
        ;; register init
        ldi     f_data_exist, $00
        ldi     address_low,  $00
        ldi     address_high, $00
        ldi     f_send_enabled, SEND_ENABLED
        sbi     PORTD, RTS

;;;======================================================================
;;; Main routine
;;;======================================================================        
main:
        sbrs    f_data_exist, 0 ; test whether command exist
        rjmp    main
        ldi     f_data_exist, 0
        ;;
        cpi     rx_data, 'u'
        breq    reset
        cpi     rx_data, 'a'
        breq    set_address
        cpi     rx_data, 'r'
        breq    read_rom
        cpi     rx_data, 'p'
        breq    read_rom_with_address_increment
        cpi     rx_data, 'b'
        breq    read_rom_bulk
        cpi     rx_data, 'w'
        breq    write_rom
        ;;
command_end:
        rjmp    main

;;;======================================================================
;;; INSTRUCTION
;;;======================================================================
;;;----------------------------------------------------------------------
;;; set address
;;;----------------------------------------------------------------------
set_address:
a_lp1:  sbrs    f_data_exist, 0         ; waiting for 1st argument
        rjmp    a_lp1
        ldi     f_data_exist, 0
        mov     address_high, rx_data
        out     PORTC, address_high
        ;; 
a_lp2:  sbrs    f_data_exist, 0         ; waiting for 2nd argument
        rjmp    a_lp2
        ldi     f_data_exist, 0
        mov     address_low, rx_data
        out     PORTA, address_low
        rjmp    command_end        
;;;----------------------------------------------------------------------
;;; read ROM
;;;----------------------------------------------------------------------
read_rom:
        cbi     PORTD, CS
        rcall   wait
        cbi     PORTD, RD
        rcall   wait
        in      tx_data, PinB
        rcall   send_data
        sbi     PORTD, RD
        sbi     PORTD, CS
        rjmp    command_end
;;;----------------------------------------------------------------------
;;; read ROM with address increment
;;;----------------------------------------------------------------------
read_rom_with_address_increment:
        cbi     PORTD, CS
        rcall   wait
        cbi     PORTD, RD
        rcall   wait
        in      tx_data, PinB
        rcall   send_data
        sbi     PORTD, RD
        sbi     PORTD, CS
        ;; address increment
        adiw    address_low, 1
        out     PORTA, address_low
        out     PORTC, address_high
        rjmp    command_end
;;;----------------------------------------------------------------------
;;; bulk reading of ROM  (this routine is not sophisticated; verbose...)
;;;----------------------------------------------------------------------
read_rom_bulk:
        ldi     b_counter, $ff
b_lp:   cbi     PORTD, CS
        rcall   wait
        cbi     PORTD, RD
        rcall   wait
        in      tx_data, PinB
        rcall   send_data
        sbi     PORTD, RD
        sbi     PORTD, CS
        ;;
        adiw    address_low, 1
        out     PORTA, address_low
        out     PORTC, address_high
        dec     b_counter
        brne    b_lp
        ;;
        cbi     PORTD, CS
        rcall   wait
        cbi     PORTD, RD
        rcall   wait
        in      tx_data, PinB
        rcall   send_data
        sbi     PORTD, RD
        sbi     PORTD, CS
        ;; 
        adiw    address_low, 1
        out     PORTA, address_low
        out     PORTC, address_high
        rjmp    command_end
;;;----------------------------------------------------------------------
;;; write ROM
;;;----------------------------------------------------------------------
write_rom:
w_lp:   sbrs    f_data_exist, 0 ; waiting for argument
        rjmp    w_lp
        ldi     f_data_exist, 0
        ;; 
        ldi     temp, $ff       ; PORTB -- 8bit data bus (output)
        out     DDRB, temp
        out     PORTB, rx_data
        rcall   wait
        cbi     PORTD, CS
        rcall   wait
        cbi     PORTD, WR
        rcall   wait
        rcall   wait
        sbi     PORTD, WR
        sbi     PORTD, CS
        ldi     temp, $00       ; PORTB -- 8bit data bus (input)
        out     DDRB, temp
        ldi     temp, $ff       ; PORTB pull up
        out     PORTB, temp
        rjmp    command_end

;;;======================================================================
;;; Subroutine
;;;======================================================================        
;;;----------------------------------------------------------------------
;;; UART send data
;;;----------------------------------------------------------------------
send_data:
        ;sbis    PinD, CTS                     
        ;rjmp    send_data
        ;ori     f_send_enabled, CTS_SEND_OK
        cpi     f_send_enabled, SEND_ENABLED
        brne    send_data
        out     UDR, tx_data
        ldi     f_send_enabled, SEND_DISABLED   ; send disabled
        ret
;;;----------------------------------------------------------------------
;;; UART send enable (interrupt routine)
;;;----------------------------------------------------------------------
send_enable:
        in      status, SREG
        ldi     f_send_enabled, SEND_ENABLED
        out     SREG, status
        reti
;;;----------------------------------------------------------------------
;;; UART receive complete (interrupt routine)
;;;----------------------------------------------------------------------
receive_complete:
        cbi     PORTD, RTS
        in      status, SREG
        in      rx_data, UDR    ; receive data
        ldi     f_data_exist, 1
        out     SREG, status
        sbi     PORTD, RTS
        reti
;;;----------------------------------------------------------------------
;;; wait routine
;;;----------------------------------------------------------------------
wait:   ret
